import re
import logging
from .consts import frequencies

"""
Rules are the base class for the very basic checks. They ensure that a type exists before more comlicated checks are run.

The important thing when writing an error for a rule is the error is the *opposite* of what the rule does.
"""


class Rule:
    def __init__(self, name: str, funct, error="") -> object:
        self.name = name  # the message printed out when the rule fails
        self.funct = funct
        self.error = error

    def run(self, x) -> bool:
        print("running rule {} on {}".format(self.name, x))
        logging.debug("running rule {} on {}".format(self.name, x))
        return self.funct(x)


no_error = Rule("there is an error when querying for the company",
                lambda x: x and (type(x) != dict or "error" not in x))

type_str = Rule("Type is string", lambda x: x is not None and type(x) == str)
type_int = Rule("Type is integer", lambda x: x is not None and type(x) == int)
type_bool = Rule("Type is boolean", lambda x: x is not None and type(x) == bool)  # check not None first
type_dct = Rule("Type is dictionary", lambda x: x is not None and type(x) == dict)
type_lst = Rule("Type is list", lambda x: x is not None and type(x) == list)
type_float = Rule("Type is float", lambda x: x is not None and type(x) in [int, float])

type_url = Rule("Is valid url (must have https://www.)",
                lambda x: x is not None and len(re.findall(r"^https://www\.(.+)\.(.+)", str(x))) > 0)

type_https = Rule("Is valid domain (must have https://)",
                  lambda x: x is not None and len(re.findall(r"^https://(.+)\.(.+)", str(x))) > 0)

type_domain = Rule("Is valid domain",
                   lambda x: x is not None and len(re.findall(r"(.+)\.(.+)", str(x))) > 0)

type_date = Rule("Is valid date", lambda x: x is not None and x)  # TODO: update this!

type_ticker = Rule("Is valid ticker, must have at least 1 letter and if a dot then at least 1 letter after ",
                   lambda x: x is not None and re.findall(r"^(\w+)((\.)\w+)?", str(x)))

non_empty_field = Rule("Is not None and not empty", lambda x: x is not None and x)

not_none = Rule("Is not None", lambda x: x is not None)

frequency_name = Rule("Is a valid frequency name",
                      lambda x: x is not None and x in frequencies)


def has_field(field_name):  # uses string name (field in x)
    return Rule("Has field `{}`".format(field_name),
                lambda x: x is not None and field_name in x,
                "does not have field `{}`".format(field_name))


def has_non_empty_field(field):  # uses field - not the name
    return Rule("Has field `{}`".format(field.name),
                lambda x: field is not None and field and (type(field) == list or
                                                           (field.value is not None and field.value)))


def in_field(field_names):  # uses actual field
    return Rule("Is within the fields `{}`".format(field_names),
                lambda x: x is not None and x in field_names)


def len_eq(number):
    return Rule("Has length `{}`".format(number),
                lambda x: x is not None and len(x) == number)


def len_gt(number):
    return Rule("Has length greater than `{}`".format(number),
                lambda x: x is not None and len(x) > number)


def len_lt(number):
    return Rule("Has length less than `{}`".format(number),
                lambda x: x is not None and len(x) < number)


def gte(number):
    return Rule("Is greater than or equal to `{}`".format(number),
                lambda x: x is not None and x >= number)


def lte(number):
    return Rule("Is less than of equal to `{}`".format(number), lambda x: x is not None and x <= number)


def matches_url(field_name):
    return Rule("Has a url for `{}`".format(field_name),
                lambda x: x is not None and matches_url_funct(field_name, x))


def matches_url_funct(field, url):
    return len(re.findall(r"https://(www\.)?{name}\.com.*".format(name=field.name), str(url))) > 0



from . import rule
from .field import Field, StringField
from .consts import id_string, name_string


class CompanyField(Field):
    def __init__(self):
        super().__init__()
        self.name = "company"
        self.value = {}

        self.id = StringField(id_string)
        self.company_name = StringField(name_string)
        self.fields = [self.id]
        self.rules = [rule.has_field(x.name) for x in self.fields]

    def set(self, value):
        self.value = value
        self.run_rules()
        if self.errors:  # checks for field names in value (type: dict)
            return self.errors
        self.set_fields(self.fields)
        return self.errors



import logging
from . import rule
from .error import Error
from .field import Field
from .company import CompanyField
from .timeseries_metadata import TimeseriesMetadataField
from .data_list import DataList


class TimeseriesCheck(Field):
    def __init__(self):
        """
        :param metadata_names:
        :param rules: list of Rule objects
        """
        super().__init__()
        self.name = "Timeseries Check"
        self.value = {}

        self.metadata = TimeseriesMetadataField()
        self.company = CompanyField()
        self.data = DataList()
        self.reference = ""  # TODO
        self.api_call = ""  # TODO
        self.fields = [self.metadata, self.company, self.data]  # will add fields when needed for use case
        self.rules = [rule.has_field(x.name) for x in self.fields]

    def set(self, value):
        self.value = value
        self.run_rules()
        if self.errors:  # returns any errors
            return self.errors
        self.metadata.rules = []  # default
        self.set_fields(self.fields)
        logging.debug("TimeseriesCheck: data field is set as {}".format(self.data))
        return self.errors

    def fill_in_error(self, error, new_check):
        error.timeseries_code = self.metadata.id.value
        error.check_name = new_check.name
        error.company = self.company.id.value
        error.reference = self.reference
        error.api_call = self.api_call
        self.error(error)

    def run_check(self, new_check, metadata_names=[], other_data = None):
        """
        Returns errors only from that check
        :param other_data:
        :param new_check:
        :param metadata_names:
        :return:
        """
        self.metadata.rules = [rule.has_field(name) for name in metadata_names]
        errors = self.validate_field(self.metadata)  # automatically adds new errors to self.errors
        if errors:  # new errors from check
            return
        error = new_check.run(self.data.value, self.metadata, other_data)
        if error is None:
            return
        if type(error) == Error and not error.empty():
            self.fill_in_error(error, new_check)
        if type(error) == list and error:
            for err in error:  # allows functions to return a list of Error objects
                if not err.empty():
                    self.fill_in_error(err, new_check)
        return

    def error(self, err: Error):
        logging.error(err)  # ERROR:root:error message
        self.errors.append(err)
        return

    def error_many(self, error_lst: list):
        if error_lst:
            self.errors += error_lst
        return

